#include <cpu.h>
#include <memory.h>
#include <stdint.h>

#define IOREDTBL(irq) (0x10 + (irq)*2)

union iored {
  struct {
    uint8_t vector;
    uint8_t flags;
    uint8_t mask;
    uint32_t reserved;
    uint8_t target;
  }__attribute__((packed));
  struct {
    uint32_t l;
    uint32_t h;
  }__attribute__((packed));
  uint64_t val;
};

static uint32_t ioapic_read(int reg) {
  uint32_t volatile *ioregsel = P2V(ioapic.addr);
  uint32_t volatile *iowin = P2V(ioapic.addr + 0x10);
  *ioregsel = reg;
  return *iowin;
}
static void ioapic_write(int reg, uint32_t value) {
  uint32_t volatile *ioregsel = P2V(ioapic.addr);
  uint32_t volatile *iowin = P2V(ioapic.addr + 0x10);
  *ioregsel = reg;
  *iowin = value;
}

static uint64_t ioapic_read_redirection(int irq) {
  union iored retval;
  retval.l = ioapic_read(IOREDTBL(irq));
  retval.h = ioapic_read(IOREDTBL(irq)+1);
  return retval.val;
}
static void ioapic_write_redirection(int irq, uint64_t val) {
  union iored value;
  value.val = val;
  ioapic_write(IOREDTBL(irq), value.l);
  ioapic_write(IOREDTBL(irq) + 1, value.h);
}

void irq_ack() {
    APIC_REG(APIC_EOI) = 0;
}

void irq_mask(int irq) {
  union iored iored;
  iored.val = ioapic_read_redirection(irq_redirects[irq]);
  iored.mask |= 0x1;
  ioapic_write_redirection(irq_redirects[irq], iored.val);
}
void irq_unmask(int irq) {
  union iored iored;
  iored.val = ioapic_read_redirection(irq_redirects[irq]);
  iored.mask &= ~0x1;
  ioapic_write_redirection(irq_redirects[irq], iored.val);
}

void ioapic_init() {
  // Map ioapic offset into kernel memory
  vmm_set_page(kernel_P4,
      (uintptr_t)P2V(ioapic.addr),
      ioapic.addr,
      PAGE_PRESENT | PAGE_WRITE | PAGE_GLOBAL
  );

  union iored iored;
  for(int i = 0; i < 24; i++) {
    iored.val = ioapic_read_redirection(i);
    iored.vector = IRQ_BASE+i;
    iored.flags = 0x0;
    iored.mask |= 1;
    iored.target = 0x0;
    ioapic_write_redirection(i, iored.val);
  }
}